import UIKit
import VisionKit

final class DocumentPreviewViewController: UIViewController {
    
    
    struct Constants {
        static let reuseIdentifier = String(describing: DocumentPreviewViewController.self)
        static let storyboardName = "DocumentPreview"
    }
    
    static func buildViewController() -> DocumentPreviewViewController {
        let controller = UIStoryboard(name: Constants.storyboardName,
                                      bundle: .main).instantiateViewController(withIdentifier: Constants.reuseIdentifier)
        return controller as! DocumentPreviewViewController
    }
    
    @IBOutlet private weak var pdfView: CustomPDFView!
    
    var file: File?
    private var scannedItem: ScannedItem!

    override func viewDidLoad() {
        super.viewDidLoad()
        
        // Do any additional setup after loading the view.
        pdfView.accessibilityIdentifier = "pdfView"
        navigationItem.title = file?.displayName
        pdfView.setDocument(file?.pdfDocument)
    }
    
    @IBAction func printTapped(_ sender: UIBarButtonItem) {
        guard let documentURL = file?.fileURL else { return }
        
        let printController = UIPrintInteractionController.shared
        printController.delegate = self
        if UIPrintInteractionController.canPrint(documentURL) {
            let printInfo = UIPrintInfo(dictionary: nil)
            printInfo.outputType = .photo
            printController.printInfo = printInfo
            printController.showsNumberOfCopies = true
        }
        
        printController.printingItems = [documentURL]
        printController.present(from: view.frame, in: view, animated: true, completionHandler: nil)
    }
    
    @IBAction func addPageTapped(_ sender: UIBarButtonItem) {
        let alertController = UIAlertController(title: nil,
                                                message: nil, preferredStyle: .actionSheet)
        
        let cameraAction = UIAlertAction(title: "Camera".localized, style: .default, handler: { action in
            self.openCamera()
        })
        
        let galleryAction = UIAlertAction(title: "Gallery".localized, style: .default, handler: { action in
            self.openGallery()
        })
        
        alertController.addAction(cameraAction)
        cameraAction.setValue(UIImage(systemName: "camera"), forKey: "image")
        cameraAction.setValue(CATextLayerAlignmentMode.left, forKey: "titleTextAlignment")
        
        alertController.addAction(galleryAction)
        galleryAction.setValue(UIImage(systemName: "photo"), forKey: "image")
        galleryAction.setValue(CATextLayerAlignmentMode.left, forKey: "titleTextAlignment")
        
        alertController.addAction(UIAlertAction(title: "Cancel".localized, style: .cancel, handler: nil))
        alertController.popoverPresentationController?.barButtonItem = sender
        present(alertController, animated: true)
    }
    
    @IBAction func shareTapped(_ sender: UIBarButtonItem) {
        guard let documentURL = file?.fileURL else { return }
        let controller = UIActivityViewController(activityItems: [documentURL], applicationActivities: nil)
        controller.popoverPresentationController?.barButtonItem = sender
        present(controller, animated: true)
    }
    
    @IBAction func drawTapped(_ sender: UIBarButtonItem) {
        let previewController = PencilKitViewController()
        previewController.dataSource = self
        previewController.delegate = self
        previewController.modalPresentationStyle = .fullScreen
        present(previewController, animated: true, completion: nil)
    }
    
    @IBAction func deleteTapped(_ sender: UIBarButtonItem) {
        let alertController = UIAlertController(title: "Delete".localized,
                                                message: "Are you sure?".localized, preferredStyle: .actionSheet)
        
        if pdfView.document?.pageCount == 1 {
            let currentPageAction = UIAlertAction(title: "Delete document".localized, style: .default, handler: { action in
                guard let documentURL = self.file?.fileURL, let pdfView = self.pdfView else { return }
                if pdfView.document?.pageCount == 1 {
                    try? FileManager.default.removeItem(at: documentURL)
                    self.navigationController?.popViewController(animated: true)
                } else {
                    PDFManager.deletePage(at: pdfView.getCurrentPageIndex(), pdfDocument: pdfView.document,
                                          documentURL: self.file?.fileURL, pdfView: pdfView)
                }
            })
            
            alertController.addAction(currentPageAction)
        } else {
            let currentPageAction = UIAlertAction(title: "Current page".localized, style: .default, handler: { action in
                guard let documentURL = self.file?.fileURL, let pdfView = self.pdfView else { return }
                if pdfView.document?.pageCount == 1 {
                    try? FileManager.default.removeItem(at: documentURL)
                    self.navigationController?.popViewController(animated: true)
                } else {
                    PDFManager.deletePage(at: pdfView.getCurrentPageIndex(), pdfDocument: pdfView.document,
                                          documentURL: self.file?.fileURL, pdfView: pdfView)
                }
            })
            
            let entireDocumentAction = UIAlertAction(title: "Entire document".localized, style: .default, handler: { action in
                guard let documentURL = self.file?.fileURL else { return }
                try? FileManager.default.removeItem(at: documentURL)
                self.navigationController?.popViewController(animated: true)
            })
            
            alertController.addAction(currentPageAction)
            alertController.addAction(entireDocumentAction)
        }
        
        alertController.addAction(UIAlertAction(title: "Cancel".localized, style: .cancel, handler: nil))
        alertController.popoverPresentationController?.barButtonItem = sender
        present(alertController, animated: true)
    }
}

// MARK: - UIPrintInteractionControllerDelegate

extension DocumentPreviewViewController: UIPrintInteractionControllerDelegate {
    
    func printInteractionControllerParentViewController(_ printInteractionController: UIPrintInteractionController) -> UIViewController? {
        return self.navigationController
    }
}

extension DocumentPreviewViewController {
    private func openCamera() {
        let scannerOptions = ImageScannerOptions(scanMultipleItems: true,
                                                 allowAutoScan: false,
                                                 allowTapToFocus: false,
                                                 defaultColorRenderOption:.color)
        let scannerViewController = ImageScannerController(options: scannerOptions)
        scannerViewController.imageScannerDelegate = self
        self.present(scannerViewController, animated: true)
    }
    
    private func openGallery() {
        guard UIImagePickerController.isSourceTypeAvailable(UIImagePickerController.SourceType.savedPhotosAlbum) else { return }
        let imagePickerController = UIImagePickerController()
        imagePickerController.delegate = self
        
        imagePickerController.sourceType = UIImagePickerController.SourceType.savedPhotosAlbum
        imagePickerController.allowsEditing = UserDefaults.standard.isPhotoEditigOn
        present(imagePickerController, animated: true, completion: nil)
    }
}

/// Tells the delegate that the user picked a still image.
extension DocumentPreviewViewController: UINavigationControllerDelegate, UIImagePickerControllerDelegate {
    func imagePickerController(_ picker: UIImagePickerController,
                               didFinishPickingMediaWithInfo info: [UIImagePickerController.InfoKey : Any]) {
        picker.dismiss(animated: true) { [weak self] in
            guard let self = self else { return }
            guard let image = info[UserDefaults.standard.isPhotoEditigOn ? UIImagePickerController.InfoKey.editedImage : UIImagePickerController.InfoKey.originalImage] as? UIImage else { return }
            PDFManager.appendPage(image: image, pdfView: self.pdfView, documentURL: self.file?.fileURL)
        }
    }
}

extension DocumentPreviewViewController: ImageScannerControllerDelegate {
    func imageScannerController(_ scanner: ImageScannerController, didFinishWithSession session: MultiPageScanSession) {
        scanner.dismiss(animated: true) { [weak self] in
            guard let self = self else { return }
            for index in 0..<session.scannedItems.count {
                if let url = session.scannedItems[index].renderedImage {
                    PDFManager.appendPage(imageURL: url, pdfView: self.pdfView, documentURL: self.file?.fileURL)
                }
                }
            }
        }
    func imageScannerControllerDidCancel(_ scanner: ImageScannerController) {
        scanner.dismiss(animated: true)
    }
    
    func imageScannerController(_ scanner: ImageScannerController, didFailWithError error: Error) {
        scanner.dismiss(animated: true)

    }
}
