//
//  EditViewController.swift
//  MiniScanner
//
//  Created by George Makhoul on 09/06/2024.
//  Copyright © 2024 AppsNectar. All rights reserved.
//

import UIKit
import LNExtensionExecutor
import NBBottomSheet
import QuickLook

public protocol EditViewControllerDelegate: NSObjectProtocol {
    func editViewController(_ editViewController: EditViewController, finished session: MultiPageScanSession)
}

public class EditViewController: UIViewController {
    // MARK: - Outlets
    @IBOutlet weak var banner: BannerView!
    @IBOutlet weak var pageIndex: UILabel!
    @IBOutlet weak var continueButton: UIButton!
    
    // MARK: - Properties
    private var scanSession:MultiPageScanSession?
    private var pages:Array<ScannedPageViewController> = []
    weak public var delegate:EditViewControllerDelegate?
    private var selectedIndex: Int = 0
    var payload: [Any] = []
    private var sharedFolder: AppConfigurator.Folder? = AppConfigurator.Folder(name: "Shared", savedName: "shared", isSelected: true)
    var selectedFolder: AppConfigurator.Folder?
    private var localFileManager: LocalFileManager?
    private var fileURL: URL?
    
    lazy private var activityIndicator: UIActivityIndicatorView = {
        let activityIndicator = UIActivityIndicatorView(style: .large)
        activityIndicator.color = .white
        activityIndicator.hidesWhenStopped = true
        activityIndicator.translatesAutoresizingMaskIntoConstraints = false
        return activityIndicator
    }()
    
    // MARK: - LifeCycle
    public override func viewDidLoad() {
        super.viewDidLoad()
        if let session = scanSession {
            banner.set(session: session)
            banner.delegate = self
            updatePageLabel()
        }
        view.addSubview(activityIndicator)
        continueButton.backgroundColor = .mainBlue
        continueButton.setTitle("Continue", for: .normal)
        continueButton.setTitleColor(.white, for: .normal)
        continueButton.layer.cornerRadius = 10
        localFileManager = LocalFileManager()

    }
    
    @IBAction func backIndex(_ sender: UIButton) {
        if banner.pagerView.currentIndex != 0 {
            banner.pagerView.scrollToItem(at: banner.pagerView.currentIndex - 1, animated: true)
            selectedIndex = banner.pagerView.currentIndex - 1
            self.updatePageLabel()
        }
    }
    
    func updatePageLabel() {
        if let session = scanSession {
            if let url = session.scannedItems[selectedIndex].renderedImage {
                payload = [url]
            }
            pageIndex.set(text: "\(selectedIndex + 1)/\(session.scannedItems.count)", color: .mainText, font: .regular(15))
            pageIndex.textAlignment = .center
        }
    }
    
    @IBAction func nextIndex(_ sender: UIButton) {
        if banner.pagerView.currentIndex != (scanSession?.scannedItems.count ?? 0) - 1 {
            banner.pagerView.scrollToItem(at: banner.pagerView.currentIndex + 1, animated: true)
            selectedIndex = banner.pagerView.currentIndex + 1
        }
        updatePageLabel()
    }
    
    @IBAction func continueTapped(_ sender: UIButton) {
        let popup = ShareSheetViewController()
        popup.session = self.scanSession
        popup.delegate = self
        popup.selectedFolder = self.selectedFolder
        let configuration = NBBottomSheetConfiguration(animationDuration: 0.4, sheetSize: .fixed(322))
        
        let bottomSheetController = NBBottomSheetController(configuration: configuration)
            bottomSheetController.present(popup, on: self)
    }
    
    public init(scanSession: MultiPageScanSession){
        self.scanSession = scanSession
        super.init(nibName: nil, bundle: nil)
    }
    
    required init?(coder aDecoder: NSCoder) {
        scanSession = MultiPageScanSession()
        fatalError("init(coder:) should not be called for this class")
    }
    
    private func directShareWith(extensionBundleIdentifier: String) async {
        do {
            let executor = try LNExtensionExecutor(extensionBundleIdentifier: extensionBundleIdentifier)
            let (completed, returnItems) = try await executor.execute(withActivityItems: payload, on: self)
            print("completed: \(completed) return items: \(returnItems)")
            
        } catch(let error) {
            print("error: \(error.localizedDescription)")
        }
    }
    
    func printPDF(from url: URL) {
        let printInfo = UIPrintInfo(dictionary:nil)
        printInfo.jobName = url.lastPathComponent
        printInfo.outputType = .general
        
        let printController = UIPrintInteractionController.shared
        printController.printInfo = printInfo
        
        let pdfData = try? Data(contentsOf: url)
        if pdfData != nil {
            printController.printingItem = pdfData
        }
        
        printController.present(animated: true, completionHandler: nil)
    }
    
    @IBAction func whatsappShare(_ sender: UIButton) {
        activityIndicator.startAnimating()
        DispatchQueue.main.async {
            
            Task {
                await self.directShareWith(extensionBundleIdentifier: ShareOptions.whatsapp)
            }
            self.activityIndicator.stopAnimating()
        }
    }
    @IBAction func telegramTapped(_ sender: UIButton) {
        activityIndicator.startAnimating()
        DispatchQueue.main.async {
            
            Task {
                await self.directShareWith(extensionBundleIdentifier: ShareOptions.telegram)
            }
            self.activityIndicator.stopAnimating()
        }
    }
    @IBAction func gmailTapped(_ sender: UIButton) {
        activityIndicator.startAnimating()
        DispatchQueue.main.async {
            Task {
                await self.directShareWith(extensionBundleIdentifier: ShareOptions.gmail)
                self.activityIndicator.stopAnimating()
            }
        }
    }
    @IBAction func printTapped(_ sender: UIButton) {
        DispatchQueue.main.async {
            self.printPDF(from: self.payload[0] as! URL)
        }
    }
    @IBAction func moreTapped(_ sender: UIButton) {
        DispatchQueue.main.async {
            self.activityIndicator.startAnimating()
            let shareController = UIActivityViewController(activityItems: self.payload, applicationActivities: nil)
            self.present(shareController, animated: true, completion: nil)
            self.activityIndicator.stopAnimating()
        }

    }
}

// MARK: - BannerViewDelegate
extension EditViewController: BannerViewDelegate {
    func preview(index: Int) {
        let previewController = PreviewImageViewController()
        if let image = scanSession?.scannedItems[index].renderedImage?.retrieveImage() {
            previewController.image = image
        }
        present(previewController, animated: true)
    }
    
    func delete(index: Int) {
        let alertController = UIAlertController(title: "Alert", message: "Are you sure you want to delete this image?", preferredStyle: .alert)

        let okAction = UIAlertAction(title: "Yes", style: .default) { (action) in
            self.scanSession?.remove(index: index)
            if index != 0 {
                self.selectedIndex = index - 1
            }
            self.banner.scanSession = self.scanSession
            self.banner.pagerView.reloadData()
            if self.scanSession?.scannedItems.count != 0 {
                self.updatePageLabel()
            } else {
                self.dismiss(animated: true) {
                    self.navigationController?.popViewController(animated: true)
                    return
                }
            }
            self.dismiss(animated: true)
        }

        let cancelAction = UIAlertAction(title: "Cancel", style: .cancel) { (action) in
            self.dismiss(animated: true)
        }

        alertController.addAction(okAction)
        alertController.addAction(cancelAction)
        self.present(alertController, animated: true, completion: nil)
    }
    
    func updateSelectedIndex(index: Int) {
        selectedIndex = index
        updatePageLabel()
    }
    
    private func saveFileAndShare(with: String, name: String) {
        activityIndicator.startAnimating()

        var images = [URL]()
        for index in 0..<(scanSession?.scannedItems.count ?? 0) {
            if let url = scanSession?.scannedItems[index].renderedImage {
                images.append(url)
            }
        }
        print("images: \(images)")
        if let folder = sharedFolder {
            PDFManager.createMultiPDFPageAndReturn(from: images, localFileManager: localFileManager, folder: folder, name: name, { document in
                self.payload = [document]
                Task {
                    await self.directShareWith(extensionBundleIdentifier: with)
                }
            })
        }
    }
}

// MARK: - ShareSheetViewControllerDelegate
extension EditViewController: ShareSheetViewControllerDelegate {
    func whatsappTapped(name: String) {
        DispatchQueue.main.async {
            self.saveFileAndShare(with: ShareOptions.whatsapp, name: name)
        }
    }
    
    func telegramTapped(name: String) {
        DispatchQueue.main.async {
            self.saveFileAndShare(with: ShareOptions.telegram, name: name)
        }
    }
    
    func airdropTapped(name: String) {
        DispatchQueue.main.async {
            self.saveFileAndShare(with: ShareOptions.gmail, name: name)
        }

    }
    
    func printTapped() {
        var images = [URL]()
        for index in 0..<(scanSession?.scannedItems.count ?? 0) {
            if let url = scanSession?.scannedItems[index].renderedImage {
                images.append(url)
            }
        }
        print("images: \(images)")
        if let folder = sharedFolder {
            PDFManager.createMultiPDFPageAndReturn(from: images, localFileManager: localFileManager, folder: folder, { document in
                self.printPDF(from: document)
            })
        }
        
    }
    
    func moreTapped(name: String) {
        activityIndicator.startAnimating()
        var images = [URL]()
        for index in 0..<(scanSession?.scannedItems.count ?? 0) {
            if let url = scanSession?.scannedItems[index].renderedImage {
                images.append(url)
            }
        }
        print("images: \(images)")
        if let folder = sharedFolder {
            PDFManager.createMultiPDFPageAndReturn(from: images, localFileManager: localFileManager, folder: folder, name: name, { document in
                self.payload = [document]
                let shareController = UIActivityViewController(activityItems: self.payload, applicationActivities: nil)
                self.present(shareController, animated: true, completion: nil)
                self.activityIndicator.stopAnimating()


            })
        }
    }
    
    func saveTapped(name: String) {
        activityIndicator.startAnimating()
        var images = [URL]()
        for index in 0..<(scanSession?.scannedItems.count ?? 0) {
            if let url = scanSession?.scannedItems[index].originalImage {
                images.append(url)
            }
        }
        print("images: \(images)")
         if let folder = selectedFolder {
            PDFManager.createMultiPDFPage(from: images, localFileManager: localFileManager, folder: folder, name: name, {
                self.activityIndicator.stopAnimating()
                self.navigationController?.popViewController(animated: true)
            })
        }
    }
}

// MARK: - QLPreviewControllerDataSource
extension EditViewController: QLPreviewControllerDataSource {
    public func numberOfPreviewItems(in controller: QLPreviewController) -> Int {
        1
    }
    
    public func previewController(_ controller: QLPreviewController, previewItemAt index: Int) -> QLPreviewItem {
        return fileURL! as QLPreviewItem
    }
}
