//
//  ScannerViewController.swift
//  WeScan
//
//  Created by Boris Emorine on 2/8/18.
//  Copyright © 2018 WeTransfer. All rights reserved.
//

import UIKit
import AVFoundation
import NBBottomSheet
import LNExtensionExecutor

/// An enum used to know if the flashlight was toggled successfully.
enum FlashResult {
    case successful
    case notSuccessful
}

struct ShareOptions {
    static var whatsapp: String = "net.whatsapp.WhatsApp.ShareExtension"
    static var telegram: String = "ph.telegra.Telegraph.Share"
    static var gmail: String = "com.google.Gmail.ShareExtension"
}

protocol ScannerViewControllerDelegate:NSObjectProtocol{
    func scannerViewController(_ scannerViewController:ScannerViewController, reviewItems inSession:MultiPageScanSession)
    func scannerViewController(_ scannerViewController:ScannerViewController, didFail withError:Error)
    func scannerViewControllerDidCancel(_ scannerViewController:ScannerViewController)
}

public final class ScannerViewController: UIViewController {
    
    var captureSessionManager: CaptureSessionManager?
    let videoPreviewLayer = AVCaptureVideoPreviewLayer()
    weak var delegate: ScannerViewControllerDelegate?
    var focusRectangle: FocusRectangleView!
    let quadView = QuadrilateralView()
    let visualEffectView = UIVisualEffectView(effect: UIBlurEffect(style: .dark))
    var flashEnabled = false
    var deviceOrientationHelper = DeviceOrientationHelper()
    var multipageSession:MultiPageScanSession!
    var options:ImageScannerOptions!
    var payload: [Any] = []
    private var localFileManager: LocalFileManager?
    private var sharedFolder: AppConfigurator.Folder? = AppConfigurator.Folder(name: "Shared", savedName: "shared", isSelected: true)
    var selectedFolder: AppConfigurator.Folder?
    var activity: UIActivityIndicatorView?
    
    public override var prefersStatusBarHidden: Bool {
        return true
    }
    
    /// The original bar style that was set by the host app
    private var originalBarStyle: UIBarStyle?
    
    lazy private var tabbarView: CustomTabBarView = {
        let view = CustomTabBarView()
        
        view.delegate = self
        return view
    }()
    
    lazy private var shutterButton: ShutterButton = {
        let button = ShutterButton()
        button.translatesAutoresizingMaskIntoConstraints = false
        button.addTarget(self, action: #selector(captureImage(_:)), for: .touchUpInside)
        return button
    }()
    
    lazy private var cancelButton: UIButton = {
        let button = UIButton()
        button.setTitle(NSLocalizedString("wescan.scanning.cancel", tableName: nil, bundle: Bundle(for: ScannerViewController.self), value: "Cancel", comment: "The cancel button"), for: .normal)
        button.translatesAutoresizingMaskIntoConstraints = false
        button.addTarget(self, action: #selector(cancelImageScannerController), for: .touchUpInside)
        return button
    }()
    
    lazy private var counterButton: UIButton = {
        let button = UIButton()
        button.setTitle("0 >", for: .normal)
        button.translatesAutoresizingMaskIntoConstraints = false
        button.backgroundColor = UIColor.init(red: 0.0, green: 0.0, blue: 0.0, alpha: 0.5)
        button.layer.borderColor = UIColor.white.cgColor
        button.layer.borderWidth = 1.0
        button.layer.cornerRadius = 5.0
        button.setTitleColor(UIColor.white, for: .normal)
        button.isHidden = true
        button.addTarget(self, action: #selector(counterImageScannerController), for: .touchUpInside)
        return button
    }()
    
    /// A black UIView, used to quickly display a black screen when the shutter button is presseed.
    internal let blackFlashView: UIView = {
        let view = UIView()
        view.backgroundColor = UIColor(white: 0.0, alpha: 0.5)
        view.isHidden = true
        view.translatesAutoresizingMaskIntoConstraints = false
        return view
    }()
    
    lazy private var autoScanButton: UIBarButtonItem = {
        var title = NSLocalizedString("wescan.scanning.auto", tableName: nil, bundle: Bundle(for: ScannerViewController.self), value: "Auto", comment: "The auto button state")
        if !self.options.allowAutoScan{
            title = NSLocalizedString("wescan.scanning.manual", tableName: nil, bundle: Bundle(for: ScannerViewController.self), value: "Manual", comment: "The manual button state")
        }
        
        let button = UIBarButtonItem(title: title, style: .plain, target: self, action: #selector(toggleAutoScan))
        button.tintColor = .white
        
        return button
    }()
    
    lazy private var flashButton: UIBarButtonItem = {
        let flashImage = UIImage(named: "flash", in: Bundle(for: ScannerViewController.self), compatibleWith: nil)
        let flashButton = UIBarButtonItem(image: flashImage, style: .plain, target: self, action: #selector(toggleFlash))
        flashButton.tintColor = .white
        
        return flashButton
    }()
    
    lazy private var activityIndicator: UIActivityIndicatorView = {
        let activityIndicator = UIActivityIndicatorView(style: .gray)
        activityIndicator.hidesWhenStopped = true
        activityIndicator.translatesAutoresizingMaskIntoConstraints = false
        return activityIndicator
    }()
    
    // MARK: - Initializers
    
    public init(scanSession:MultiPageScanSession?, options:ImageScannerOptions?) {
        self.multipageSession = scanSession ?? MultiPageScanSession()
        self.options = options ?? ImageScannerOptions()
        super.init(nibName: nil, bundle: nil)
    }
    
    init() {
        super.init(nibName: nil, bundle: nil)
    }
    
    
    required init?(coder aDecoder: NSCoder) {
        self.multipageSession = MultiPageScanSession()
        self.options = ImageScannerOptions()
        super.init(coder: aDecoder)
    }
    
    // MARK: - Life Cycle
    
    public override func viewDidLoad() {
        super.viewDidLoad()
        
        title = nil
        localFileManager = LocalFileManager()
        setupViews()
        setupNavigationBar()
        setupConstraints()
        
        captureSessionManager = CaptureSessionManager(videoPreviewLayer: videoPreviewLayer)
        captureSessionManager?.delegate = self
        CaptureSession.current.isAutoScanEnabled = self.options.allowAutoScan
        
        originalBarStyle = navigationController?.navigationBar.barStyle
        
        NotificationCenter.default.addObserver(self, selector: #selector(subjectAreaDidChange), name: Notification.Name.AVCaptureDeviceSubjectAreaDidChange, object: nil)
    }
    
    public override func viewWillAppear(_ animated: Bool) {
        super.viewWillAppear(animated)
        setNeedsStatusBarAppearanceUpdate()
        
        CaptureSession.current.isEditing = false
        quadView.removeQuadrilateral()
        captureSessionManager?.start()
        UIApplication.shared.isIdleTimerDisabled = true
        navigationController?.navigationBar.isTranslucent = true
        navigationController?.navigationBar.setBackgroundImage(UIImage(), for: .default)        
        navigationController?.setToolbarHidden(true, animated: true)
        
        updateCounterButton()
        
        deviceOrientationHelper.startDeviceOrientationNotifier { (deviceOrientation) in
            self.orientationChanged(deviceOrientation: deviceOrientation)
        }
    }
    
    public override func viewDidDisappear(_ animated: Bool) {
        deviceOrientationHelper.stopDeviceOrientationNotifier()
        super.viewDidDisappear(true)
    }
    
    public override func viewDidLayoutSubviews() {
        super.viewDidLayoutSubviews()
        
        videoPreviewLayer.frame = view.layer.bounds
        
        let statusBarHeight = UIApplication.shared.statusBarFrame.size.height
        let visualEffectRect = self.navigationController?.navigationBar.bounds.insetBy(dx: 0, dy: -(statusBarHeight)).offsetBy(dx: 0, dy: -statusBarHeight)
        
        visualEffectView.frame = visualEffectRect ?? CGRect.zero
    }
    
    public override func viewWillDisappear(_ animated: Bool) {
        super.viewWillDisappear(animated)
        UIApplication.shared.isIdleTimerDisabled = false
        guard let device = AVCaptureDevice.default(for: AVMediaType.video) else { return }
        if device.torchMode == .on {
            toggleFlash()
        }
    }
    
    // MARK: - Setups
    
    private func setupViews() {
        view.layer.addSublayer(videoPreviewLayer)
        quadView.translatesAutoresizingMaskIntoConstraints = false
        quadView.editable = false
        view.addSubview(quadView)
        view.addSubview(tabbarView)
        view.addSubview(activityIndicator)
        view.addSubview(blackFlashView)
        view.addSubview(tabbarView)
    }
    
    private func setupNavigationBar() {
        navigationItem.setLeftBarButton(flashButton, animated: false)
        navigationItem.setRightBarButton(autoScanButton, animated: false)
        
        if UIImagePickerController.isFlashAvailable(for: .rear) == false {
            let flashOffImage = UIImage(named: "flashUnavailable", in: Bundle(for: ScannerViewController.self), compatibleWith: nil)
            flashButton.image = flashOffImage
            flashButton.tintColor = UIColor.lightGray
        }
    }
    
    private func setupConstraints() {
        var quadViewConstraints = [NSLayoutConstraint]()
        var cancelButtonConstraints = [NSLayoutConstraint]()
        var shutterButtonConstraints = [NSLayoutConstraint]()
        var activityIndicatorConstraints = [NSLayoutConstraint]()
        var counterButtonConstraints = [NSLayoutConstraint]()
        var blackFlashViewConstraints = [NSLayoutConstraint]()
        var tabbarViewConstraints = [NSLayoutConstraint]()
        
        quadViewConstraints = [
            quadView.topAnchor.constraint(equalTo: view.topAnchor),
            view.bottomAnchor.constraint(equalTo: quadView.bottomAnchor),
            view.trailingAnchor.constraint(equalTo: quadView.trailingAnchor),
            quadView.leadingAnchor.constraint(equalTo: view.leadingAnchor)
        ]
        
        shutterButtonConstraints = [
            shutterButton.centerXAnchor.constraint(equalTo: view.centerXAnchor),
            shutterButton.widthAnchor.constraint(equalToConstant: 65.0),
            shutterButton.heightAnchor.constraint(equalToConstant: 65.0)
        ]
        
        tabbarViewConstraints = [
            tabbarView.leadingAnchor.constraint(equalTo: view.leadingAnchor, constant: 0),
            tabbarView.trailingAnchor.constraint(equalTo: view.trailingAnchor, constant: 0),
            tabbarView.bottomAnchor.constraint(equalTo: view.bottomAnchor, constant: 0),
            tabbarView.heightAnchor.constraint(equalToConstant: 225)
        ]
        
        activityIndicatorConstraints = [
            activityIndicator.centerXAnchor.constraint(equalTo: view.centerXAnchor),
            activityIndicator.centerYAnchor.constraint(equalTo: view.centerYAnchor)
        ]
        
        blackFlashViewConstraints = [
            blackFlashView.topAnchor.constraint(equalTo: view.topAnchor),
            blackFlashView.leadingAnchor.constraint(equalTo: view.leadingAnchor),
            view.bottomAnchor.constraint(equalTo: blackFlashView.bottomAnchor),
            view.trailingAnchor.constraint(equalTo: blackFlashView.trailingAnchor)
        ]
        
        if #available(iOS 11.0, *) {
            cancelButtonConstraints = [
                cancelButton.leftAnchor.constraint(equalTo: view.safeAreaLayoutGuide.leftAnchor, constant: 24.0),
                view.safeAreaLayoutGuide.bottomAnchor.constraint(equalTo: cancelButton.bottomAnchor, constant: (65.0 / 2) - 10.0)
            ]
            
            counterButtonConstraints = [
                counterButton.centerYAnchor.constraint(equalTo: shutterButton.centerYAnchor),
                counterButton.rightAnchor.constraint(equalTo: view.safeAreaLayoutGuide.rightAnchor, constant: -24.0),
                counterButton.widthAnchor.constraint(equalToConstant: 44.0),
                counterButton.heightAnchor.constraint(equalToConstant: 44.0)
            ]
            
            let shutterButtonBottomConstraint = view.safeAreaLayoutGuide.bottomAnchor.constraint(equalTo: shutterButton.bottomAnchor, constant: 8.0)
            shutterButtonConstraints.append(shutterButtonBottomConstraint)
        } else {
            cancelButtonConstraints = [
                cancelButton.leftAnchor.constraint(equalTo: view.leftAnchor, constant: 24.0),
                view.bottomAnchor.constraint(equalTo: cancelButton.bottomAnchor, constant: (65.0 / 2) - 10.0)
            ]
            
            let shutterButtonBottomConstraint = view.bottomAnchor.constraint(equalTo: shutterButton.bottomAnchor, constant: 8.0)
            shutterButtonConstraints.append(shutterButtonBottomConstraint)
        }
        tabbarView.translatesAutoresizingMaskIntoConstraints = false

        NSLayoutConstraint.activate(quadViewConstraints + activityIndicatorConstraints + blackFlashViewConstraints + tabbarViewConstraints)
    }
    
    private func flashToBlack() {
        view.bringSubviewToFront(blackFlashView)
        blackFlashView.isHidden = false
        let flashDuration = DispatchTime.now() + 0.05
        DispatchQueue.main.asyncAfter(deadline: flashDuration) {
            self.blackFlashView.isHidden = true
        }
    }
    
    // MARK: - Tap to Focus
    
    /// Called when the AVCaptureDevice detects that the subject area has changed significantly. When it's called, we reset the focus so the camera is no longer out of focus.
    @objc private func subjectAreaDidChange() {
        /// Reset the focus and exposure back to automatic
        do {
            try CaptureSession.current.resetFocusToAuto()
        } catch {
            let error = ImageScannerControllerError.inputDevice
            guard let captureSessionManager = captureSessionManager else { return }
            captureSessionManager.delegate?.captureSessionManager(captureSessionManager, didFailWithError: error)
            return
        }
        
        /// Remove the focus rectangle if one exists
        CaptureSession.current.removeFocusRectangleIfNeeded(focusRectangle, animated: true)
    }
    
    public override func touchesBegan(_ touches: Set<UITouch>, with event: UIEvent?) {
        super.touchesBegan(touches, with: event)
        
        if self.options.allowTapToFocus{
            guard  let touch = touches.first else { return }
            let touchPoint = touch.location(in: view)
            let convertedTouchPoint: CGPoint = videoPreviewLayer.captureDevicePointConverted(fromLayerPoint: touchPoint)
            
            CaptureSession.current.removeFocusRectangleIfNeeded(focusRectangle, animated: false)
            
            focusRectangle = FocusRectangleView(touchPoint: touchPoint)
            view.addSubview(focusRectangle)
            
            do {
                try CaptureSession.current.setFocusPointToTapPoint(convertedTouchPoint)
            } catch {
                let error = ImageScannerControllerError.inputDevice
                guard let captureSessionManager = captureSessionManager else { return }
                captureSessionManager.delegate?.captureSessionManager(captureSessionManager, didFailWithError: error)
                return
            }
        }
    }
    
    private func updateCounterButton(){
        tabbarView.updateScannedView(session: self.multipageSession)
    }
    
    // MARK: Rotation methods
    private func orientationChanged(deviceOrientation: UIDeviceOrientation) {
        print("Orientation changed: \(self.deviceOrientationHelper.currentDeviceOrientation.rawValue)")
    }
    
    private func getCurrentRotationAngle()->Double{
        switch self.deviceOrientationHelper.currentDeviceOrientation {
        case .landscapeRight:
            return -90.0
        case .landscapeLeft:
            return 90.0
        case .portrait:
            return 0.0
        case .portraitUpsideDown:
            return 180.0
        default:
            return 0.0
        }
    }
    
    // MARK: - Actions
    
    @objc private func captureImage(_ sender: UIButton) {
        self.flashToBlack()
        shutterButton.isUserInteractionEnabled = false
        captureSessionManager?.capturePhoto()
    }
    
    @objc private func toggleAutoScan() {
        if CaptureSession.current.isAutoScanEnabled {
            CaptureSession.current.isAutoScanEnabled = false
            autoScanButton.title = NSLocalizedString("wescan.scanning.manual", tableName: nil, bundle: Bundle(for: ScannerViewController.self), value: "Manual", comment: "The manual button state")
        } else {
            CaptureSession.current.isAutoScanEnabled = true
            autoScanButton.title = NSLocalizedString("wescan.scanning.auto", tableName: nil, bundle: Bundle(for: ScannerViewController.self), value: "Auto", comment: "The auto button state")
        }
    }
    
    @objc private func toggleFlash() {
        let state = CaptureSession.current.toggleFlash()
        
        let flashImage = UIImage(named: "flash", in: Bundle(for: ScannerViewController.self), compatibleWith: nil)
        let flashOffImage = UIImage(named: "flashUnavailable", in: Bundle(for: ScannerViewController.self), compatibleWith: nil)
        
        switch state {
        case .on:
            flashEnabled = true
            flashButton.image = flashImage
            flashButton.tintColor = .yellow
        case .off:
            flashEnabled = false
            flashButton.image = flashImage
            flashButton.tintColor = .white
        case .unknown, .unavailable:
            flashEnabled = false
            flashButton.image = flashOffImage
            flashButton.tintColor = UIColor.lightGray
        }
    }
    
    @objc private func cancelImageScannerController() {
        self.navigationController?.popViewController(animated: true)
    }
    
    @objc private func counterImageScannerController(){
        self.delegate?.scannerViewController(self, reviewItems: self.multipageSession)
    }
    
    private func saveFileAndShare(with: String, name: String) {
        activityIndicator.startAnimating()

        var images = [URL]()
        for index in 0..<multipageSession.scannedItems.count {
            if let url = multipageSession.scannedItems[index].renderedImage {
                images.append(url)
            }
        }
        print("images: \(images)")
        if let folder = sharedFolder {
            PDFManager.createMultiPDFPageAndReturn(from: images, localFileManager: localFileManager, folder: folder, name: name, { document in
                self.payload = [document]
                Task {
                    await self.directShareWith(extensionBundleIdentifier: with)
                }
            })
        }
    }
    
    private func directShareWith(extensionBundleIdentifier: String) async {
        do {
            let executor = try LNExtensionExecutor(extensionBundleIdentifier: extensionBundleIdentifier)
            let (completed, returnItems) = try await executor.execute(withActivityItems: payload, on: self)
            print("completed: \(completed) return items: \(returnItems)")
            activityIndicator.stopAnimating()

        } catch(let error) {
            print("error: \(error.localizedDescription)")
        }
    }
    
}

extension ScannerViewController: RectangleDetectionDelegateProtocol {
    func captureSessionManager(_ captureSessionManager: CaptureSessionManager, didFailWithError error: Error) {
        
        activityIndicator.stopAnimating()
        
        self.delegate?.scannerViewController(self, didFail: error)
    }
    
    func didStartCapturingPicture(for captureSessionManager: CaptureSessionManager) {
        activityIndicator.startAnimating()
//        shutterButton.isUserInteractionEnabled = false
    }
    
    func captureSessionManager(_ captureSessionManager: CaptureSessionManager, didCapturePicture picture: UIImage, withQuad quad: Quadrilateral?) {
        activityIndicator.stopAnimating()
        
        let scannedItem = ScannedItem(originalImage:picture, quad:quad)
        scannedItem.rotation = self.getCurrentRotationAngle()
        scannedItem.colorOption = self.options.defaultColorRenderOption
        scannedItem.render { (_) in }
        self.multipageSession.add(item: scannedItem)
        self.updateCounterButton()
        self.captureSessionManager?.start()
    }
    
    func captureSessionManager(_ captureSessionManager: CaptureSessionManager, didDetectQuad quad: Quadrilateral?, _ imageSize: CGSize) {
        guard let quad = quad else {
            // If no quad has been detected, we remove the currently displayed on on the quadView.
            quadView.removeQuadrilateral()
            return
        }
        
        let portraitImageSize = CGSize(width: imageSize.height, height: imageSize.width)
        
        let scaleTransform = CGAffineTransform.scaleTransform(forSize: portraitImageSize, aspectFillInSize: quadView.bounds.size)
        let scaledImageSize = imageSize.applying(scaleTransform)
        
        let rotationTransform = CGAffineTransform(rotationAngle: CGFloat.pi / 2.0)
        
        let imageBounds = CGRect(origin: .zero, size: scaledImageSize).applying(rotationTransform)
        
        let translationTransform = CGAffineTransform.translateTransform(fromCenterOfRect: imageBounds, toCenterOfRect: quadView.bounds)
        
        let transforms = [scaleTransform, rotationTransform, translationTransform]
        
        let transformedQuad = quad.applyTransforms(transforms)
        
        quadView.drawQuadrilateral(quad: transformedQuad, animated: true)
    }
    
}

// MARK: - CustomTabBarViewDelegate
extension ScannerViewController: CustomTabBarViewDelegate {
    func scanTapped() {
        self.flashToBlack()
        captureSessionManager?.capturePhoto()
    }
    
    func fileManagerTapped() {
        self.navigationController?.popViewController(animated: false)
    }
    func editButtonTapped() {
        print(multipageSession.scannedItems.count)
        let editViewController = EditViewController(scanSession: self.multipageSession)
        editViewController.selectedFolder = self.selectedFolder
        self.navigationController?.pushViewController(editViewController, animated: true)
    }
    
    func shareTapped() {
        let popup = ShareSheetViewController()
        popup.session = self.multipageSession
        popup.delegate = self
        popup.selectedFolder = self.selectedFolder
        let configuration = NBBottomSheetConfiguration(animationDuration: 0.4, sheetSize: .fixed(322))
        
        let bottomSheetController = NBBottomSheetController(configuration: configuration)
            bottomSheetController.present(popup, on: self)
    }
    
    func printPDF(from url: URL) {
        let printInfo = UIPrintInfo(dictionary:nil)
        printInfo.jobName = url.lastPathComponent
        printInfo.outputType = .general

        let printController = UIPrintInteractionController.shared
        printController.printInfo = printInfo

        let pdfData = try? Data(contentsOf: url)
        if pdfData != nil {
            printController.printingItem = pdfData
        }

        printController.present(animated: true, completionHandler: nil)
    }
}
// MARK: - ShareSheetViewControllerDelegate
extension ScannerViewController: ShareSheetViewControllerDelegate {
    func whatsappTapped(name: String) {
        DispatchQueue.main.async {
            self.saveFileAndShare(with: ShareOptions.whatsapp, name: name)
        }
    }
    
    func telegramTapped(name: String) {
        DispatchQueue.main.async {
            self.saveFileAndShare(with: ShareOptions.telegram, name: name)
        }
    }
    
    func airdropTapped(name: String) {
        DispatchQueue.main.async {
            self.saveFileAndShare(with: ShareOptions.gmail, name: name)
        }

    }
    
    func printTapped() {
        var images = [URL]()
        for index in 0..<multipageSession.scannedItems.count {
            if let url = multipageSession.scannedItems[index].renderedImage {
                images.append(url)
            }
        }
        print("images: \(images)")
        if let folder = sharedFolder {
            PDFManager.createMultiPDFPageAndReturn(from: images, localFileManager: localFileManager, folder: folder, { document in
                self.printPDF(from: document)
            })
        }
        
    }
    
    func moreTapped(name: String) {
        activityIndicator.startAnimating()
        var images = [URL]()
        for index in 0..<multipageSession.scannedItems.count {
            if let url = multipageSession.scannedItems[index].renderedImage {
                images.append(url)
            }
        }
        print("images: \(images)")
        if let folder = sharedFolder {
            PDFManager.createMultiPDFPageAndReturn(from: images, localFileManager: localFileManager, folder: folder, name: name, { document in
                self.payload = [document]
                let shareController = UIActivityViewController(activityItems: self.payload, applicationActivities: nil)
                self.present(shareController, animated: true, completion: nil)
                self.activityIndicator.stopAnimating()


            })
        }
    }
    
    func saveTapped(name: String) {
        activityIndicator.startAnimating()
        var images = [URL]()
        for index in 0..<multipageSession.scannedItems.count {
            if let url = multipageSession.scannedItems[index].renderedImage {
                images.append(url)
            }
        }
        print("images: \(images)")
        if let folder = selectedFolder {
            PDFManager.createMultiPDFPage(from: images, localFileManager: localFileManager, folder: folder, name: name, {
                self.activityIndicator.stopAnimating()
                self.navigationController?.popViewController(animated: true)
            })
        }
    }
}
